#define _GNU_SOURCE
#include "common.h"

#include <errno.h>
#include <error.h>
#include <fcntl.h>
#include <inttypes.h>
#include <limits.h>
#include <pthread.h>
#include <sched.h>
#include <signal.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

static int latency_target_fd = -1;
static int32_t latency_target_value = 0;

void (*previous_handlers[NSIG])(int);
static void (*sighand)(int);

uint64_t ts_to_uint(struct timespec t) {
  return t.tv_sec * NSEC_PER_SEC + t.tv_nsec;
}
struct timespec uint_to_ts(uint64_t t) {
  struct timespec ts;
  ts.tv_sec = t / NSEC_PER_SEC;
  ts.tv_nsec = t - (ts.tv_sec * NSEC_PER_SEC);
  return ts;
}

/* Latency trick
 * if the file /dev/cpu_dma_latency exists,
 * open it and write a zero into it. This will tell
 * the power management system not to transition to
 * a high cstate (in fact, the system acts like idle=poll)
 * When the fd to /dev/cpu_dma_latency is closed, the behavior
 * goes back to the system default.
 *
 * Documentation/power/pm_qos_interface.txt
 */
void set_latency_target(void) {
  struct stat s;
  int err;

  errno = 0;
  err = stat("/dev/cpu_dma_latency", &s);
  if (err == -1) {
    error(EXIT_FAILURE, errno, "WARN: stat /dev/cpu_dma_latency failed");
    return;
  }

  errno = 0;
  latency_target_fd = open("/dev/cpu_dma_latency", O_RDWR);
  if (latency_target_fd == -1) {
    error(EXIT_FAILURE, errno, "WARN: open /dev/cpu_dma_latency");
    return;
  }

  errno = 0;
  err = write(latency_target_fd, &latency_target_value, 4);
  if (err < 1) {
    error(EXIT_FAILURE, errno, "# error setting cpu_dma_latency to %d!",
          latency_target_value);
    close(latency_target_fd);
    return;
  }
  printf("# /dev/cpu_dma_latency set to %dus\n", latency_target_value);
}

void add_ns(struct timespec *t, uint64_t ns) {
  t->tv_nsec += ns;

  while (t->tv_nsec >= ((int64_t)NSEC_PER_SEC)) {
    t->tv_sec += 1;
    t->tv_nsec -= NSEC_PER_SEC;
  }
}

uint64_t calcdiff_ns(struct timespec t1, struct timespec t2) {
  uint64_t diff;
  diff = NSEC_PER_SEC * (uint64_t)((int)t1.tv_sec - (int)t2.tv_sec);
  diff += ((int)t1.tv_nsec - (int)t2.tv_nsec);
  return diff;
}

int64_t _max_(int64_t a, int64_t b) { return a > b ? a : b; }
int64_t _min_(int64_t a, int64_t b) { return a < b ? a : b; }

int histogram_min(uint64_t *histogram, int max_value) {
  int ret = max_value;
  for (int i = max_value; i >= 0; i--) ret = histogram[i] ? i : ret;
  return ret;
}
int histogram_max(uint64_t *histogram, int max_value) {
  int ret = 0;
  for (int i = 0; i <= max_value; i++) ret = histogram[i] ? i : ret;
  return ret;
}

static void sighand_wrapper(int sig) {
  // If we get un unexpected signal, report it, if not print the histogram
  if (sig == SIGINT || sig == SIGTERM)
    (*sighand)(sig);  // Will print the histogram
  else
    printf("Uknown signal interrupt: %s (%d)\n", strsignal(sig), sig);

  // Execute the default handler
  if (previous_handlers[sig] == SIG_DFL) {
    signal(sig, SIG_DFL);
    raise(sig);
  } else if (previous_handlers[sig] == SIG_IGN) {
    return;
  } else {
    (*previous_handlers[sig])(sig);
  }
}

void init_signals(void (*_sighand)(int)) {
  sighand = _sighand;

  for (int i = 0; i < NSIG; i++) signal(i, sighand_wrapper);
}

static int get_bit(uint64_t t, int i) {
  return (t & (((uint64_t) 1) << i)) > 0;
}
void encode(uint64_t t, char * b) {
  for(int i = 0; i < 64; i++)
    b[i] = get_bit(t, i) * 0xff;
}
uint64_t decode(char * b) {
  uint64_t ret = 0;
  for(int i = 0; i < 64; i++) {
    int sum_bits = 0;
    for(int j = 0; j < 8; j++)
      sum_bits += (b[i] & (1 << j)) > 0;
    //printf("Sum: %d (%x)\n", sum_bits, b[i]);
    ret |= (sum_bits >= 4) * (((uint64_t) 1) << i);
  }
  return ret;
}
